@@ -2,6 +2,7 @@
 # but adapted to use relay, automatic field detection and some code adjustments
 
 import collections
+import collections.abc
 import itertools
 
 from django.core.exceptions import (
@@ -13,6 +14,7 @@ from django.db import (
     models,
     transaction,
 )
+from django.db.models.fields import NOT_PROVIDED
 from django.db.models.fields.reverse_related import (
     ManyToOneRel,
     ManyToManyRel,
@@ -22,12 +24,16 @@ from graphene.relay.mutation import ClientIDMutation
 from graphene.types.mutation import MutationOptions
 from graphene.types.utils import yank_fields_from_attrs
 from graphene_django.registry import get_global_registry
-from graphene_django.converter import convert_django_field_with_choices
+from graphene_django.converter import (
+    convert_django_field_with_choices,
+    get_choices,
+)
 from graphene.utils.str_converters import to_camel_case
 from graphql.error import GraphQLError
 
 from .exceptions import PermissionDenied
 from .models import GuardedModel
+from .schema import get_field_schema
 from .types import (
     MutationErrorType,
     UploadType,
@@ -43,6 +49,17 @@ from .utils import (
 )
 
 _registry = get_global_registry()
+input_schema_registry = {}
+
+
+def _update_dict_nested(d, u):
+    for k, v in u.items():
+        if isinstance(v, collections.abc.Mapping):
+            d[k] = _update_dict_nested(d.get(k, {}), v)
+        else:
+            d[k] = v
+
+    return d
 
 
 def _get_model_name(model):
@@ -87,8 +104,7 @@ def _get_validation_errors(validation_error):
 
 def _get_fields(model, only_fields, exclude_fields, required_fields):
     fields = [
-        (field.name, field)
-        for field in sorted(list(model._meta.fields + model._meta.many_to_many))
+        (field.name, field) for field in sorted(list(model._meta.fields + model._meta.many_to_many))
     ]
 
     # Only add the field as input field if the foreign key (reverse side)
@@ -106,30 +122,32 @@ def _get_fields(model, only_fields, exclude_fields, required_fields):
 
     ret = collections.OrderedDict()
     for name, field in fields:
-        if ((only_fields and name not in only_fields) or
-                name in exclude_fields or
-                str(name).endswith('+') or
-                name in ['created_at', 'updated_at', 'archived_at']):
+        if (
+            (only_fields and name not in only_fields)
+            or name in exclude_fields
+            or str(name).endswith("+")
+            or name in ["created_at", "updated_at", "archived_at"]
+        ):
             continue
 
-        if name == 'id':
-            ret[name] = graphene.ID(
+        if name == "id":
+            f = graphene.ID(
                 description="The ID of the object.",
             )
         elif isinstance(field, models.FileField):
-            ret[name] = UploadType(
+            f = UploadType(
                 description=field.help_text,
             )
         elif isinstance(field, models.BooleanField):
-            ret[name] = graphene.Boolean(
+            f = graphene.Boolean(
                 description=field.help_text,
             )
         elif isinstance(field, (models.ForeignKey, models.OneToOneField)):
-            ret[name] = graphene.ID(
+            f = graphene.ID(
                 description=field.help_text,
             )
         elif isinstance(field, models.ManyToManyField):
-            ret[name] = graphene.List(
+            f = graphene.List(
                 graphene.ID,
                 description=field.help_text,
             )
@@ -139,43 +157,74 @@ def _get_fields(model, only_fields, exclude_fields, required_fields):
             if isinstance(field, ManyToOneRel) and not reverse_rel_include and not only_fields:
                 continue
 
-            ret[name] = graphene.List(
+            f = graphene.List(
                 graphene.ID,
-                description='Set list of {0}'.format(
+                description="Set list of {0}".format(
                     field.related_model._meta.verbose_name_plural,
                 ),
             )
         else:
-            ret[name] = convert_django_field_with_choices(field, _registry)
+            f = convert_django_field_with_choices(field, _registry)
 
         if required_fields is not None:
-            ret[name].kwargs['required'] = name in required_fields
+            f.kwargs["required"] = name in required_fields
         else:
             if isinstance(field, (ManyToOneRel, ManyToManyRel)):
-                ret[name].kwargs['required'] = not field.null
+                f.kwargs["required"] = not field.null
             else:
-                ret[name].kwargs['required'] = not field.blank
+                f.kwargs["required"] = not field.blank and field.default is NOT_PROVIDED
+
+        if getattr(field, "choices", None):
+            items = field.choices
+            if isinstance(items, dict):
+                items = items.items()
+
+            choices = []
+            for (original_v, label), (n, value, desc) in zip(items, get_choices(items)):
+                choices.append(
+                    {
+                        "label": label,
+                        "name": n,
+                        "value": value,
+                    }
+                )
+        else:
+            choices = None
+
+        s = get_field_schema(field)
+        s.update(
+            {
+                "name": name,
+                # FIXME: Get verbose_name and help_text for m2m
+                "label": getattr(field, "verbose_name", None),
+                "help_text": getattr(field, "help_text", None),
+                "max_length": getattr(field, "max_length", None),
+                "choices": choices,
+            }
+        )
+
+        ret[name] = {
+            "field": f,
+            "schema": s,
+        }
 
     return ret
 
 
 def _is_list_of_ids(field):
-    return (
-        isinstance(field.type, graphene.List) and
-        field.type.of_type == graphene.ID
-    )
+    return isinstance(field.type, graphene.List) and field.type.of_type == graphene.ID
 
 
 def _is_id_field(field):
     return (
-        field.type == graphene.ID or
-        isinstance(field.type, graphene.NonNull) and
-        field.type.of_type == graphene.ID
+        field.type == graphene.ID
+        or isinstance(field.type, graphene.NonNull)
+        and field.type.of_type == graphene.ID
     )
 
 
 def _is_upload_field(field):
-    t = getattr(field.type, 'of_type', field.type)
+    t = getattr(field.type, "of_type", field.type)
     return t == UploadType
 
 
@@ -191,6 +240,9 @@ class BaseMutationOptions(MutationOptions):
     #: If we should allow unauthenticated users to do this mutation
     allow_unauthenticated = False
 
+    #: The input schema for the schema query
+    input_schema = None
+
 
 class BaseMutation(ClientIDMutation):
     """Base mutation enchanced with permission checking and relay id handling."""
@@ -205,21 +257,33 @@ class BaseMutation(ClientIDMutation):
     )
 
     @classmethod
-    def __init_subclass_with_meta__(cls, permissions=None,
-                                    permissions_any=True,
-                                    allow_unauthenticated=False,
-                                    _meta=None, **kwargs):
+    def __init_subclass_with_meta__(
+        cls,
+        permissions=None,
+        permissions_any=True,
+        allow_unauthenticated=False,
+        input_schema=None,
+        _meta=None,
+        **kwargs,
+    ):
         if not _meta:
             _meta = BaseMutationOptions(cls)
 
         _meta.permissions = permissions or []
         _meta.permissions_any = permissions_any
         _meta.allow_unauthenticated = allow_unauthenticated
+        _meta.input_schema = input_schema or {}
 
         super().__init_subclass_with_meta__(_meta=_meta, **kwargs)
 
+        iname = cls.Input._meta.name
+        input_schema_registry[iname] = {
+            "input_object": iname,
+            "fields": list(_meta.input_schema.values()),
+        }
+
     @classmethod
-    def get_node(cls, info, node_id, field='id', only_type=None):
+    def get_node(cls, info, node_id, field="id", only_type=None):
         """Get the node object given a relay global id."""
         if not node_id:
             return None
@@ -230,9 +294,7 @@ class BaseMutation(ClientIDMutation):
             raise ValidationError({field: str(e)})
         else:
             if node is None:  # pragma: no cover
-                raise ValidationError(
-                    {field: "Couldn't resolve to a node: {}".format(node_id)}
-                )
+                raise ValidationError({field: "Couldn't resolve to a node: {}".format(node_id)})
 
         return node
 
@@ -259,8 +321,7 @@ class BaseMutation(ClientIDMutation):
         if not cls._meta.permissions:
             return True
 
-        return check_perms(user, cls._meta.permissions,
-                           any_perm=cls._meta.permissions_any)
+        return check_perms(user, cls._meta.permissions, any_perm=cls._meta.permissions_any)
 
     @classmethod
     def mutate_and_get_payload(cls, root, info, **data):
@@ -334,13 +395,19 @@ class BaseModelMutation(BaseMutation):
         abstract = True
 
     @classmethod
-    def __init_subclass_with_meta__(cls, model=None,
-                                    object_permissions=None,
-                                    object_permissions_any=True,
-                                    return_field_name=None,
-                                    required_fields=None,
-                                    exclude_fields=None, only_fields=None,
-                                    _meta=None, **kwargs):
+    def __init_subclass_with_meta__(
+        cls,
+        model=None,
+        object_permissions=None,
+        object_permissions_any=True,
+        return_field_name=None,
+        required_fields=None,
+        exclude_fields=None,
+        only_fields=None,
+        input_schema=None,
+        _meta=None,
+        **kwargs,
+    ):
         if not model:  # pragma: no cover
             raise ImproperlyConfigured("model is required for ModelMutation")
         if not _meta:
@@ -351,9 +418,16 @@ class BaseModelMutation(BaseMutation):
         if not return_field_name:
             return_field_name = _get_model_name(model)
 
-        input_fields = _get_fields(model, only_fields, exclude_fields,
-                                   required_fields)
-        input_fields = yank_fields_from_attrs(input_fields, _as=graphene.InputField)
+        fdata = _get_fields(model, only_fields, exclude_fields, required_fields)
+        input_fields = yank_fields_from_attrs(
+            {k: v["field"] for k, v in fdata.items()},
+            _as=graphene.InputField,
+        )
+
+        input_schema = _update_dict_nested(
+            {k: v["schema"] for k, v in fdata.items()},
+            input_schema or {},
+        )
 
         fields = _get_output_fields(model, return_field_name)
 
@@ -368,6 +442,7 @@ class BaseModelMutation(BaseMutation):
         super().__init_subclass_with_meta__(
             _meta=_meta,
             input_fields=input_fields,
+            input_schema=input_schema,
             **kwargs,
         )
 
@@ -477,8 +552,8 @@ class ModelOperationMutation(BaseModelMutation):
     @classmethod
     def __init_subclass_with_meta__(cls, **kwargs):
         super().__init_subclass_with_meta__(
-            only_fields=['id'],
-            required_fields=['id'],
+            only_fields=["id"],
+            required_fields=["id"],
             **kwargs,
         )
 
@@ -499,9 +574,7 @@ class ModelMutation(BaseModelMutation):
         opts = instance._meta
 
         for f in opts.fields:
-            if (not f.editable or
-                    isinstance(f, models.AutoField) or
-                    f.name not in cleaned_data):
+            if not f.editable or isinstance(f, models.AutoField) or f.name not in cleaned_data:
                 continue
 
             data = cleaned_data[f.name]
@@ -523,7 +596,7 @@ class ModelMutation(BaseModelMutation):
     def clean_instance(cls, instance, clean_input):
         """Validate the instance by calling its `.full_clean()` method."""
         try:
-            instance.full_clean()
+            instance.full_clean(exclude=cls._meta.exclude_fields)
         except ValidationError as e:
             if e.error_dict:
                 raise e
@@ -566,7 +639,7 @@ class ModelMutation(BaseModelMutation):
         Create or update the instance, based on the existence of the
         `id` attribute in the input data and save it.
         """
-        obj_id = data.get('id')
+        obj_id = data.get("id")
         if obj_id:
             checked_permissions = True
             instance = cls.get_instance(info, obj_id)
@@ -591,7 +664,7 @@ class ModelMutation(BaseModelMutation):
                 if d is not None:
                     target_field = getattr(instance, f.related_name or f.name + "_set")
                     target_field.set(d)
-            elif hasattr(f, 'save_form_data'):
+            elif hasattr(f, "save_form_data"):
                 d = cleaned_input.get(f.name, None)
                 if d is not None:
                     f.save_form_data(instance, d)
@@ -618,9 +691,9 @@ class ModelCreateMutation(ModelMutation):
 
     @classmethod
     def __init_subclass_with_meta__(cls, **kwargs):
-        exclude_fields = kwargs.pop('exclude_fields', []) or []
-        if 'id' not in exclude_fields:
-            exclude_fields.append('id')
+        exclude_fields = kwargs.pop("exclude_fields", []) or []
+        if "id" not in exclude_fields:
+            exclude_fields.append("id")
         super().__init_subclass_with_meta__(
             exclude_fields=exclude_fields,
             **kwargs,
@@ -639,11 +712,11 @@ class ModelUpdateMutation(ModelMutation):
 
     @classmethod
     def __init_subclass_with_meta__(cls, **kwargs):
-        if 'only_fields' in kwargs and 'id' not in kwargs['only_fields']:
-            kwargs['only_fields'].insert(0, 'id')
-        required_fields = kwargs.pop('required_fields', []) or []
-        if 'id' not in required_fields:
-            required_fields.insert(0, 'id')
+        if "only_fields" in kwargs and "id" not in kwargs["only_fields"]:
+            kwargs["only_fields"].insert(0, "id")
+        required_fields = kwargs.pop("required_fields", []) or []
+        if "id" not in required_fields:
+            required_fields.insert(0, "id")
         super().__init_subclass_with_meta__(
             required_fields=required_fields,
             **kwargs,
@@ -664,7 +737,7 @@ class ModelDeleteMutation(ModelOperationMutation):
         Delete the instance from the database given its `id` attribute
         in the input data.
         """
-        instance = cls.get_instance(info, data.get('id'))
+        instance = cls.get_instance(info, data.get("id"))
 
         db_id = instance.id
         cls.delete(info, instance)
